import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Hashtable;

/**
 * Der Serverprozess fr den Spielserver.
 * 
 * @author Q11
 * @version 2015-07-08
 */
public class ServerProzess implements Runnable
{
    /*
     * Der ServerSocket, auf dem ich laufe.
     */
    private ServerSocket serverSocket;

    /*
     * Eine Tabelle, die zu jedem Nick (Anmeldenamen) den
     * zugehrigen Clientantwortprozess enthlt.
     */
    private Hashtable <String, ClientAntwortProzess> prozess;

    /*
     * Eine Tabelle, die zu jedem Nick (Anmeldenamen) das
     * zugehoerige Passwort enthaelt.
     */
    private Hashtable <String, String> passwort;

    /*
     * Eine Tabelle, die zu jedem Nick (Anmeldenamen) den
     * Nick der Wunschpartnerin fr das nchste Spiel enthlt.
     */
    private Hashtable <String, String> wunschpartnerin;

    /*
     * Zwei Tabellen, die zu jedem Nick (Anmeldenamen) die
     * Gren fr das nchste Spiel enthalten.
     */
    private Hashtable <String, Integer> wunschgroesseX;
    private Hashtable <String, Integer> wunschgroesseY;

    /*
     * Eine Tabelle, die zu jedem Nick (Anmeldenamen) sein
     * aktuelles Spiel enthlt.
     */
    private Hashtable <String, Spiel> spiel;

    /*
     * Eine Tabelle, die zu jedem Nick (Anmeldenamen) seine
     * aktuelle Punktzahl enthlt.
     */
    private Hashtable <String, Integer> punkte;

    /*
     * Eine Tabelle, die die zu jedem Nick (Anmeldenamen)
     * seinen aktullen Farbwert von rot enthlt.
     */
    private Hashtable <String, Integer> rot;

    /*
     * Eine Tabelle, die die zu jedem Nick (Anmeldenamen)
     * seinen aktullen Farbwert von grn enthlt.
     */
    private Hashtable <String, Integer> gruen;

    /*
     * Eine Tabelle, die die zu jedem Nick (Anmeldenamen)
     * seinen aktullen Farbwert von blau enthlt.
     */
    private Hashtable <String, Integer> blau;

    /**
     * Konstruktor.
     * Erzeuge einen Serverprozess auf dem angegebenen ServerSocket.
     * Beginne mit leeren Tabellen.
     */
    public ServerProzess (ServerSocket s)
    {
        serverSocket = s;

        prozess         = new Hashtable <String, ClientAntwortProzess> ();
        passwort        = new Hashtable <String, String> ();
        punkte          = new Hashtable <String, Integer> ();
        wunschgroesseX  = new Hashtable <String, Integer> ();
        wunschgroesseY  = new Hashtable <String, Integer> ();
        spiel           = new Hashtable <String, Spiel> ();
        wunschpartnerin = new Hashtable <String, String> ();
        rot             = new Hashtable <String, Integer> ();
        gruen           = new Hashtable <String, Integer> ();
        blau            = new Hashtable <String, Integer> ();
    }

    /**
     * Ablaufplan.
     * Warte auf neue Clientverbindungen. Starte dabei fr jeden
     * sich anmeldenden Client einen neuen ClientAntwortProzess.
     */
    public void run ()
    {
        try
        {
            while (true)
            {
                Socket clientSocket = serverSocket.accept ();
                new Thread (new ClientAntwortProzess (clientSocket, this)).start();
            }
        }
        catch (IOException fehler)
        {
            System.err.println ("Serverprozess scheitert.");
            System.exit (1);
        }
    }

    /**
     * Sende die angegebene Nachricht an alle.
     */
    public synchronized void sendeAnAlle (String nachricht)
    {
        for (ClientAntwortProzess c: prozess.values ())
        {
            c.sende (nachricht);
        }
    }

    /**
     * Sende die angegebene Nachricht an die angegebene Person.
     */
    public synchronized boolean sendeAn (String name, String nachricht, ClientAntwortProzess absender)
    {
        if (prozess.containsKey (name))
        {
            ClientAntwortProzess empfaenger = prozess.get (name);
            empfaenger.sende (nachricht);

            return true;
        }
        else
        {
            absender.sende ("Niemand heit " + name + ".");

            return false;
        }
    }

    /**
     * Pruefe, ob ein Nick existiert.
     */
    public synchronized boolean pruefeNick (String name)
    {
        return passwort.containsKey (name);
    }

    /**
     * Erstelle oder ndere den Nick eines Clientantwortprozesses.
     */
    public synchronized boolean aendereNick (String alterName, String neuerName, ClientAntwortProzess absender)
    {
        if (prozess.containsKey (neuerName))
        {
            absender.sende ("Jemand heit schon " + neuerName + ".");

            return false;
        }
        else if (pruefeName (neuerName) == false)
        {
            absender.sende ("Der Name darf nur Gross-, Kleinbuchstaben und Ziffern enthalten!");
            
            return false;
        }
        else
        {
            prozess.put (neuerName, absender);

            /*
             * Falls der Name schon existierte ...
             */
            if (alterName != null)
            {
                /*
                 * ... ndere seine Identitt in allen Tabellen.
                 */
                if (punkte.containsKey (alterName))
                {
                    punkte.put (neuerName, punkte.get (alterName));
                }

                if (passwort.containsKey (alterName))
                {
                    passwort.put (neuerName, passwort.get (alterName));
                }

                if (rot.containsKey (alterName))
                {
                    rot.put (neuerName, rot.get (alterName));
                }

                if (gruen.containsKey (alterName))
                {
                    gruen.put (neuerName, gruen.get (alterName));
                }

                if (blau.containsKey (alterName))
                {
                    blau.put (neuerName, blau.get (alterName));
                }

                if (spiel.containsKey (alterName))
                {
                    spiel.put (neuerName, spiel.get (alterName));
                }

                if (wunschpartnerin.containsKey (alterName))
                {
                    wunschpartnerin.put (neuerName, wunschpartnerin.get (alterName));
                }

                if (wunschgroesseX.containsKey (alterName))
                {
                    wunschgroesseX.put (neuerName, wunschgroesseX.get (alterName));
                }

                if (wunschgroesseY.containsKey (alterName))
                {
                    wunschgroesseY.put (neuerName, wunschgroesseY.get (alterName));
                }

                for (String nutzerin: Collections.list (wunschpartnerin.keys ()))
                {
                    if (wunschpartnerin.get (nutzerin).equals (alterName))
                    {
                        wunschpartnerin.remove (nutzerin);
                        wunschpartnerin.put (nutzerin, neuerName);
                    }
                }
                
                prozess.remove (alterName);
                punkte.remove (alterName);
                spiel.remove (alterName);
                wunschpartnerin.remove (alterName);
                wunschgroesseX.remove (alterName);
                wunschgroesseY.remove (alterName);
                passwort.remove (alterName);
                rot.remove (alterName);
                gruen.remove (alterName);
                blau.remove (alterName);

                /*
                 * Mache die Namensnderung bekannt.
                 */
                sendeAnAlle (alterName + " heit ab jetzt " + neuerName + ".");
            }
            else
            {
                /*
                 * Andernfalls nimm die neue Benutzerin auf.
                 */
                if (!punkte.containsKey (neuerName))
                {
                    punkte.put (neuerName, 0);
                }
                
                sendeAnAlle (neuerName + " ist online.");
            }

            return true;
        }

    }

    /**
     * Prfe, ob der angegebene Name zulssig ist (also nur
     * aus Klein- und Grobuchstaben sowie Ziffern besteht).
     */  
    private boolean pruefeName (String name)
    {
        char zeichen;
        boolean erlaubt = true;

        for (int i = 0; i < name.length (); i++)
        {
            zeichen = name.charAt (i);

            if ('a' <= zeichen && zeichen <= 'z')
            {
            }
            else if ('A' <= zeichen && zeichen <= 'Z')
            {
            }
            else if ('0' <= zeichen && zeichen <= '9')
            {
            }
            else
            {
                erlaubt = false;
            }
        }

        return erlaubt;
    }

    /**
     * Pruefe das Kennwort der Benutzerin mit dem angegebenen Namen.
     */
    public synchronized boolean pruefePasswort (String name, String kennwortEingabe)
    {
        return passwort.get (name).equals (kennwortEingabe);
    }

    /**
     * Gib der Benutzerin mit dem angegebenen Namen das angegebene Kennwort.
     */
    public synchronized void setzePasswort (String name, String kennwort)
    {
        passwort.put (name, kennwort);
    }

    /**
     * Erstelle oder ndere die Farbe der Benutzerin name.
     */
    public synchronized boolean aendereFarbe (String name, int r, int g, int b)
    {
        boolean erlaubt = true;

        for (String nutzerin: Collections.list (rot.keys ()))
        {
            if (!nutzerin.equals(name))
            {
                if (!farbabstand(r, g, b, rot.get(nutzerin), gruen.get(nutzerin), blau.get(nutzerin)))
                { 
                    erlaubt = false;
                    prozess.get (name).sende (nutzerin + " hat eine zu hnliche Farbe!" + " rot:" + rot.get(nutzerin) + " grn:" + gruen.get(nutzerin) + " blau:" + blau.get(nutzerin));
                }
            }
        }

        if (!farbabstand(r, g, b, 127))
        { 
            erlaubt = false;
            prozess.get (name).sende ("Das Spielfeld hat eine zu hnliche Farbe!" + " rot:127" + " grn:127" + " blau:127" );
        }

        if (!farbabstand(r, g, b, 255))
        { 
            erlaubt = false;
            prozess.get (name).sende ("Deine Farbe darf nicht wei sein!" + " wei:" + " rot:255" + " grn:255" + " blau:255" );
        }

        if (erlaubt)
        { 
            rot.put (name, r);
            gruen.put (name, g);
            blau.put (name, b);
        }

        return erlaubt;
    }

    /**
     * Prfe, ob der Abstand zwischen den Farben (r1, g1, b1) und
     * (r2, g2. b2) gro genug ist, dass man sie unterscheiden kann.
     */
    private boolean farbabstand (int r1, int g1, int b1, int r2, int g2, int b2)
    {
        int rotabstand   = Math.abs (r1 - r2);
        int gruenabstand = Math.abs (g1 - g2);
        int blauabstand  = Math.abs (b1 - b2);

        if  (rotabstand*rotabstand + gruenabstand*gruenabstand + blauabstand*blauabstand >= 3000)
        {
            return true;
        }
        else
        { 
            return false;
        }
    }

    /**
     * Pruefe, ob der Abstand der Farben (r1, g1, b1) 
     * gross genug ist, dass man sie von der Spielfeldfarbe grau
     * unterscheiden kann.
     */
    private boolean farbabstand (int r1, int g1, int b1, int grau)
    {
        int rotabstand   = Math.abs (r1 - grau);
        int gruenabstand = Math.abs (g1 - grau);
        int blauabstand  = Math.abs (b1 - grau);

        if  (rotabstand*rotabstand + gruenabstand*gruenabstand + blauabstand*blauabstand >= 3000)
        {
            return true;
        }
        else
        { 
            return false;
        }
    }

    /**
     * Melde den angegebenen Nick und seinen Clientantwortprozess ab.
     */
    public synchronized void meldeAb (String name)
    {
        prozess.remove (name);
        spiel.remove (name);
        wunschpartnerin.remove (name);

        sendeAnAlle (name + " ist offline.");
    }

    /**
     * Teile der Benutzerin name alle Nicks und ihre Herkunft mit,
     * deren Nick mit muster beginnt.
     */
    public synchronized void nenneNicks (String name, String muster)
    {
        for (String nutzerin: Collections.list (prozess.keys ()))
        {
            if (nutzerin.startsWith (muster))
            {
                prozess.get (name).sende (nutzerin + " ist verbunden von " + prozess.get (nutzerin).gibAdresse () + ".");
            }
        }
    }

    /**
     * Vermerke, dass die Spielerin sich ein Spiel mit bestimmter Spielfeldgre mit der Gegnerin wnscht.
     * Wenn der Wunsch auf Gegenseitigkeit beruht, starte ein neues Spiel.
     */
    public synchronized void wuenscheSpiel (String spielerin, String gegnerin, int groesseX, int groesseY)
    {
        /*
         * Wenn die Gegnerin existiert ...
         */
        if (prozess.containsKey (gegnerin))
        {
            /*
             * ... vermerke sie als Wunschpartnerin der Spielerin.
             */
            wunschpartnerin.put (spielerin, gegnerin);
            
            /*
             * Vermerke die Wunschgren der Spielerin.
             */
            wunschgroesseX.put (spielerin, groesseX);
            wunschgroesseY.put (spielerin, groesseY);

            /*
             * Wenn die Spielerin auch Wunschpartnerin der Gegnerin ist ...
             */
            if (spielerin.equals (wunschpartnerin.get (gegnerin)))
            {
                /*
                 * ... und die gewnschten Spielfeldgren bereinstimmen ...
                 */
                if (groesseX == wunschgroesseX.get (gegnerin) && groesseY == wunschgroesseY.get (gegnerin))
                {
                    /*
                     * ... lose die Reihenfolge aus und ...
                     */
                    double zufall = Math.random ();
                    String ersteSpielerin;
                    String zweiteSpielerin;

                    if (Math.random () < 0.5)
                    {
                        ersteSpielerin = spielerin;
                        zweiteSpielerin = gegnerin;
                    }
                    else
                    {
                        ersteSpielerin = gegnerin;
                        zweiteSpielerin = spielerin;
                    }

                    /*
                     * ... beginne ein neues Spiel.
                     */
                    Spiel neuesSpiel = new Spiel (ersteSpielerin, zweiteSpielerin, groesseX, groesseY);

                    spiel.put (spielerin, neuesSpiel);
                    spiel.put (gegnerin, neuesSpiel);

                    /*
                     * Versetze beide Kontrahentinnen in den Zustand SPIELEN.
                     */
                    prozess.get (spielerin).spiele (gegnerin);
                    prozess.get (gegnerin).spiele (spielerin);

                    /*
                     * Kndige das Spiel serverweit an.
                     */
                    sendeAnAlle ("Neues Spiel zwischen " + ersteSpielerin + " und " + zweiteSpielerin + ".");

                    /*
                     * Sende beiden Clients das Signal zum Spielstart.
                     */
                    prozess.get (spielerin).sende ("Spiele gegen " + gegnerin + " " + rot.get (gegnerin) + " " + gruen.get (gegnerin) + " " + blau.get (gegnerin) + " auf " + wunschgroesseX.get (gegnerin) + " " + wunschgroesseY.get (gegnerin));
                    prozess.get (gegnerin).sende ("Spiele gegen " + spielerin + " " + rot.get (spielerin) + " " + gruen.get (spielerin) + " " + blau.get (spielerin) + " auf " + wunschgroesseX.get (spielerin) + " " + wunschgroesseY.get (spielerin));
                    
                    prozess.get (spielerin).sende ("faerbe spieleranzeige fuer " + ersteSpielerin);
                    prozess.get (gegnerin).sende ("faerbe spieleranzeige fuer " + ersteSpielerin);
                }
            }
            /*
             * Falls die Gegnerin bereits spielt ...
             */
            else if (spiel.containsKey (gegnerin))
            {
                /*
                 * ... melde dies der Spielerin.
                 */
                prozess.get (spielerin).sende (gegnerin + " spielt bereits in einem anderen Spiel.");
            }
        }
    }

    /**
     * Spiele im eigenen Spiel an der angegebenen Linie.
     * Die richtung ("hoch" oder "quer") gibt an, ob es
     * sich um eine Hochlinie oder um eine Querlinie
     * handelt.
     */
    public synchronized void spiele (String name, String xText, String yText, String richtung)
    {
        /*
         * Wenn die Spielerin gerade ein Spiel spielt ...
         */
        if (spiel.containsKey (name))
        {
            /*
             * Ermittle dieses Spiel und die zugehrige Gegnerin.
             */
            Spiel meinSpiel = spiel.get (name);
            String gegnerin = wunschpartnerin.get (name);
            String gewinnerin;
            int x;
            int y;
            Ergebnis spielergebnis;

            /*
             * Versuche, die angegebene Linie zu frben.
             */
            try
            {
                /*
                 * Ermittle die Koordinaten der Linie.
                 */
                x = Integer.parseInt (xText);
                y = Integer.parseInt (yText);

                /*
                 * Faerbe die angegebene Linie.
                 */
                if (richtung.equals ("hoch"))
                {
                    spielergebnis = meinSpiel.faerbeHochbalken (name, x, y);
                }
                else if (richtung.equals ("quer"))
                {
                    spielergebnis = meinSpiel.faerbeQuerbalken (name, x, y);
                }
                else
                {
                    spielergebnis = Ergebnis.RICHTUNGSFEHLER;
                }

                /*
                 * Reagiere je nach Ergebnis des Spielzugs.
                 */
                switch (spielergebnis)
                {
                    /*
                     * Falls die Spielerin nicht dran war, melde dies.
                     */
                    case NICHT_DRAN:
                    prozess.get (name).sende ("Du bist nicht dran.");
                    break;

                    /*
                     * Falls die Spaltennummer ungltig war, melde dies.
                     */
                    case UNGUELTIG:
                    //prozess.get (name).sende (spalte + " is not a valid column number.");
                    break;

                    /*
                     * Falls die Richtung ungltig war, melde dies.
                     */
                    case RICHTUNGSFEHLER:
                    prozess.get (name).sende (richtung + " ist keine gltige Spielrichtung.");
                    break;

                    /*
                     * Falls die Linie schon gefrbt war, melde dies.
                     */
                    case VOLL:
                    if (richtung.equals ("hoch"))
                    {
                        prozess.get (name).sende ("Die Hochlinie " + x +" " + y + " ist schon gefrbt!");
                    }
                    else
                    {
                        prozess.get (name).sende ("Die Querlinie " + x + " " + y + " ist schon gefrbt!");
                    }
                    break;
                    
                    /*
                     * Falls das Spiel durch den Spielzug mit einem Sieg oder
                     * einem Untentschieden endet ...
                     */
                    case SPIELERIN1:
                    case SPIELERIN2:
                    case UNENTSCHIEDEN:
                    
                    /*
                     * Ermittle die Gewinnerin.
                     */
                    gewinnerin = meinSpiel.gibGewinnerin ();

                    /*
                     * Falls es keine Gewinnerin gibt ...
                     */
                    if (gewinnerin == null)
                    {
                        /*
                         * ... verknde ein Unentschieden und gib beiden
                         * Spielerinnen je einen Punkt.
                         */
                        sendeAnAlle (name + " hat unentschieden gespielt gegen " + gegnerin + ".");
                        punkte.put (name, punkte.get (name) + 1);
                        punkte.put (gegnerin, punkte.get (gegnerin) + 1);
                        prozess.get (name).sende ("Punkte durch Spiel: 1");
                        prozess.get (gegnerin).sende ("Punkte durch Spiel: 1");
                    }
                    else if (gewinnerin.equals (name))
                    {
                        sendeAnAlle (name + " hat gewonnen gegen " + gegnerin + ".");
                        punkte.put (name, punkte.get (name) + 2);
                        prozess.get (name).sende ("Punkte durch Spiel: 2");
                        prozess.get (gegnerin).sende ("Punkte durch Spiel: 0");
                    }
                    else
                    {
                        sendeAnAlle (gegnerin + " hat gewonnen gegen " + name + ".");
                        punkte.put (gegnerin, punkte.get (gegnerin) + 2);
                        prozess.get (name).sende ("Punkte durch Spiel: 0");
                        prozess.get (gegnerin).sende ("Punkte durch Spiel: 2");
                    }

                    /*
                     * Sorge dafr, dass die Kontrahentinnen nicht mehr
                     * Wunschpartnerinnen voneinander sind (denn sonst knnte
                     * eine der beiden eine Revanche erzwingen).
                     */
                    wunschpartnerin.remove (name);
                    wunschpartnerin.remove (gegnerin);
                    
                    /*
                     * Versetze beide Clients in den Zustand WERTEN.
                     */
                    prozess.get (name).werte ();
                    prozess.get (gegnerin).werte ();

                    /*
                     * Falls der Spielzug geklappt hat und das Spiel weiterluft,
                     * melde beiden Clients den Spielzug.
                     */
                    case WEITER:
                    if (richtung.equals ("hoch"))
                    {
                        prozess.get (name).sende ("moved hoch " + name + " " + x +" " + y );
                        prozess.get (gegnerin).sende ("moved hoch " + name + " " + x +" " + y );
                    }
                    else
                    { 
                        prozess.get (name).sende ("moved quer " + name + " " + x + " " + y);
                        prozess.get (gegnerin).sende ("moved quer " + name + " " + x + " " + y);
                    }

                    int x1 = meinSpiel.gibErsteFaerbungX();
                    int y1 = meinSpiel.gibErsteFaerbungY();
                    int x2 = meinSpiel.gibZweiteFaerbungX();
                    int y2 = meinSpiel.gibZweiteFaerbungY();

                    if (x1 != -1 && y1 != -1)
                    {
                        prozess.get (name).sende ("faerbe Kaestchen fuer " + name + " " + x1 + " " + y1);
                        prozess.get (gegnerin).sende ("faerbe Kaestchen fuer " + name + " " + x1 + " " + y1);
                    }

                    if (x2 != -1 && y2 != -1)
                    {
                        prozess.get (name).sende ("faerbe Kaestchen fuer " + name + " " + x2 +" " + y2);
                        prozess.get (gegnerin).sende ("faerbe Kaestchen fuer " + name + " " + x2 +" " + y2);
                    }

                    prozess.get (name).sende ("faerbe spieleranzeige fuer " + meinSpiel.gibNaechsteSpielerin ());
                    prozess.get (gegnerin).sende ("faerbe spieleranzeige fuer " + meinSpiel.gibNaechsteSpielerin ());
                    break;
                }
            }
            /*
             * Falls die Koordinaten keine Zahlen waren, melde dies.
             */
            catch (NumberFormatException fehler)
            {
                prozess.get (name).sende (xText + " und " + yText + " sind keine gltigen Koordinaten!");
            }
        }
        /*
         * Falls die Spielerin kein Spiel am Laufen hat, melde dies.
         */
        else
        {
            prozess.get (name).sende ("Du spielst kein Spiel.");
        }
    }

    /**
     * Gib im eigenen Spiel auf.
     */
    public synchronized void gibAuf (String name)
    {
        /*
         * Falls die Spielerin gerade ein Spiel spielt ...
         */
        if (spiel.containsKey (name))
        {
            /*
             * ... ermittle ihre Gegnerin.
             */
            String gegnerin = wunschpartnerin.get (name);

            /*
             * Verknde die Niederlage und gib der Gegnerin zwei Siegpunkte.
             */
            sendeAnAlle (name + " hat gegen " + gegnerin + " aufgegeben.");
            punkte.put (gegnerin, punkte.get (gegnerin) + 2);
            
            /*
             * Melde beiden Clients die Niederlage.
             */
            prozess.get (name).sende ("Punkte durch Aufgeben: 0");
            prozess.get (gegnerin).sende ("Punkte durch Aufgeben: 2");
            
            /*
             * Sorge dafr, dass die Kontrahentinnen nicht mehr
             * Wunschpartnerinnen voneinander sind (denn sonst knnte
             * eine der beiden eine Revanche erzwingen).
             */
            wunschpartnerin.remove (name);
            wunschpartnerin.remove (gegnerin);
            
            /*
             * Versetze beide Clients in den Zustand WERTEN.
             */
            prozess.get (name).werte ();
            prozess.get (gegnerin).werte ();
        }
        /*
         * Falls die Spielerin kein Spiel am Laufen hat, melde dies.
         */
        else
        {
            prozess.get (name).sende ("Du spielst kein Spiel.");
        }
    }

    /**
     * Beende das eigene Spiel.
     */
    public synchronized void beendeSpiel (String name)
    {
        /*
         * Falls die Spielerin noch ein Spiel besitzt ...
         */
        if (spiel.containsKey (name))
        {
            /*
             * ... lsche das Spiel.
             */
            spiel.remove (name);

            /*
             * Melde dem Client, dass das Spiel gelscht ist.
             */
            prozess.get (name).sende ("Fertig.");
            
            /*
             * Versetze den Client in den Zustand CHATTEN.
             */
            prozess.get (name).chatte ();
        }
        /*
         * Falls die Spielerin kein Spiel besitzt, melde dies.
         */
        else
        {
            prozess.get (name).sende ("Du spielst kein Spiel.");
        }
    }

    /**
     * Teile der Benutzerin name die Punkte aller Spielerinnen mit,
     * deren Nick mit muster beginnt.
     */
    public synchronized void nennePunkte (String name, String muster)
    {
        /*
         * Sammle die Punktnachrichten aller betroffenen Spielerinnen,
         * und zwar so, dass sie mit den rechtsbndigen, fnfstelligen
         * Punktzahlen beginnen.
         */
        ArrayList<String> punktnachricht = new ArrayList<String> ();
        
        for (String nutzerin: Collections.list (punkte.keys ()))
        {
            if (nutzerin.startsWith (muster))
            {
                int punktzahl     = punkte.get (nutzerin);
                String punktkette = "    " + punktzahl;
                
                punktkette = punktkette.substring (punktkette.length () - 5);

                if (1 == punktzahl)
                {
                    punktnachricht.add (punktkette + " Punkt hat " + nutzerin + ".");
                }
                else
                {
                    punktnachricht.add (punktkette + " Punkte hat " + nutzerin + ".");
                }
            }
        }
        
        /*
         * Sende die punktnachrichten sortiert an die Benutzerin name.
         */
        Collections.sort (punktnachricht);
        
        for (String nachricht: punktnachricht)
        {
            prozess.get (name).sende (nachricht);
        }
    }

    /**
     * Teile der Benutzerin ihre Farbe mit.
     */
    public synchronized void nenneFarbe (String name)
    {
        prozess.get (name).sende ("Farben: " + rot.get (name) + " " + gruen.get (name) + " " + blau.get (name));
    }
}