/**
 * Eine Partie "Ksekstchen".
 * 
 * @author Q11
 * @version 2015-07-08
 */
public class Spiel
{
    /*
     * Die Namen der ersten und der zweiten Spielerin.
     */
    private String spielerin1;
    private String spielerin2;

    /*
     * Die Punktzahlen der beiden Spielerinnen.
     */
    private int punkte1;
    private int punkte2;

    /*
     * Welche Spielerin ist als nchste dran?
     */
    private int naechsteSpielerin;

    /*
     * Die Querlinien, Hochlinien und Kstchen des Spielfelds.
     */
    private boolean [][] querlinie;
    private boolean [][] hochlinie;
    private int [][] kaestchen;

    /*
     * Die Gre des Spielfelds.
     */
    private int x;
    private int y;

    /*
     * Der Name der Gewinnerin.
     */
    private String gewinnerin;

    /*
     * Die Anzahl der Spielerinnen.
     */
    //private int spielerinzahl;

    /*
     * Die Koordinaten des ersten und des zweiten gefrbten
     * Kstchens (-1, falls keine Frbung vorliegt).
     */
    private int ersteFaerbungX;
    private int ersteFaerbungY;
    private int zweiteFaerbungX;
    private int zweiteFaerbungY;

    /**
     * Konstruktor.
     * Starte ein Spiel zwischen ersterName und zweiterName
     * auf einem Brett der Gre x*y.
     */
    public Spiel (String ersterName, String zweiterName, int x, int y)
    {
        /*
         * Speichere die Spielfeldgre.
         */
        this.x = x;
        this.y = y;

        /*
         * Erzeuge die Linien und die Kstchen.
         */
        querlinie = new boolean [x][y+1];
        hochlinie = new boolean [x+1][y];
        kaestchen = new int     [x][y];

        /*
         * Markiere den Spielfeldrand und vermerke dabei
         * bei jedem Randkstchen, wie viele Nachbarlinien
         * schon markiert sind.
         */
        for (int i = 0; i < x; i++)
        {
            for (int j = 0; j <= y; j++)
            {
                if (0 == j)
                {
                    querlinie[i][j] = true;
                    kaestchen[i][0]++;
                }
                
                if (y == j)
                {
                    querlinie[i][j] = true;
                    kaestchen [i][j-1]++;
                }
            }
        }

        for (int i = 0; i <= x; i++)
        {
            for (int j = 0; j < y; j++)
            {
                if (0 == i )
                {
                    hochlinie[i][j] = true;
                    kaestchen [0][j]++;
                }
                
                if ( x == i)
                {
                    hochlinie[i][j] = true;
                    kaestchen [i-1][j]++;
                }
            }
        }

        /*
         * Speichere die Namen der Spielerinnen.
         */
        spielerin1 = ersterName;
        spielerin2 = zweiterName;

        /*
         * Die erstgenannte Spielerin beginnt.
         */
        naechsteSpielerin = 1;

        /*
         * Noch hat niemand gewonnen.
         */
        punkte1= 0;
        punkte2= 0;

        gewinnerin = null;
    }

    /**
     * Frbe den Hochbalken (i|j) fr die angegebene Spielerin.
     */
    public Ergebnis faerbeHochbalken (String name, int i, int j)
    {
        /*
         * Anfangs sind keine neuen Kstchen gefrbt.
         */
        boolean gefaerbt = false;

        ersteFaerbungX = -1;
        ersteFaerbungY = -1;

        zweiteFaerbungX = -1;
        zweiteFaerbungY = -1;

        /*
         * Wenn die Spielerin dran ist ...
         */
        if ((naechsteSpielerin == 1 && name.equals (spielerin1)) || (naechsteSpielerin == 2 && name.equals (spielerin2)))
        {
            /*
             * ... und die Hochlinie schon belegt ist ...
             */
            if (hochlinie[i][j] == true)
            {
                /*
                 * ... melde dies ...
                 */
                return Ergebnis.VOLL;
            }
            /*
             * ... sonst ...
             */
            else
            {
                /*
                 * ... belege die Hochlinie.
                 */
                hochlinie[i][j] = true;

                /*
                 * Beim Kstchen links davon ist eine Randlinie mehr gefrbt.
                 */
                kaestchen[i-1][j]++;

                /*
                 * Falls insgesamt vier Randlinien gefrbt sind ...
                 */
                if (kaestchen[i-1][j] == 4)
                {
                    /*
                     * ... frbe das Kstchen.
                     */
                    ersteFaerbungX = i-1;
                    ersteFaerbungY = j;
                    gefaerbt = true;
                    gibPunkt ();
                }

                /*
                 * Beim Kstchen rechts davon ist eine Randlinie mehr gefrbt.
                 */
                kaestchen[i][j]++;

                /*
                 * Falls insgesamt vier Randlinien gefrbt sind ...
                 */
                if (kaestchen[i][j] == 4)
                {
                    /*
                     * ... frbe das Kstchen.
                     */
                    zweiteFaerbungX = i;
                    zweiteFaerbungY = j;
                    gefaerbt = true;
                    gibPunkt ();
                }

                /*
                 * Falls kein Kstchen gefrbt wurde ...
                 */
                if (!gefaerbt)
                {
                    /*
                     * ... wechsle zur anderen Spielerin.
                     */
                    naechsteSpielerin = 3 - naechsteSpielerin;
                }

                return pruefeSpielende ();
            }
        }
        /*
         * Wenn die Spielerin nicht dran ist ...
         */
        else 
        {
            /*
             * ... melde dies.
             */
            return Ergebnis.NICHT_DRAN;
        }
    }

    /**
     * Frbe den Querbalken (i|j) fr die angegebene Spielerin.
     */
    public Ergebnis faerbeQuerbalken (String name, int i, int j)
    {
        /*
         * Anfangs sind keine neuen Kstchen gefrbt.
         */
        boolean gefaerbt = false;

        ersteFaerbungX = -1;
        ersteFaerbungY = -1;

        zweiteFaerbungX = -1;
        zweiteFaerbungY = -1;

        /*
         * Wenn die Spielerin dran ist ...
         */
        if ((naechsteSpielerin == 1 && name.equals (spielerin1)) || (naechsteSpielerin == 2 && name.equals (spielerin2)))
        {
            /*
             * ... und die Querlinie schon belegt ist ...
             */
            if (querlinie[i][j] == true)
            {
                /*
                 * ... melde dies ...
                 */
                return Ergebnis.VOLL;
            }
            /*
             * ... sonst ...
             */ 
            else
            {
                /*
                 * ... belege die Querlinie.
                 */
                querlinie[i][j] = true;

                /*
                 * Beim Kstchen darber ist eine Randlinie mehr gefrbt.
                 */
                kaestchen[i][j-1]++;

                /*
                 * Falls insgesamt vier Randlinien gefrbt sind ...
                 */
                if (kaestchen[i][j-1] == 4)
                {
                    /*
                     * ... frbe das Kstchen.
                     */
                    ersteFaerbungX = i;
                    ersteFaerbungY = j-1;
                    gefaerbt = true;
                    gibPunkt ();
                }

                /*
                 * Beim Kstchen darunter ist eine Randlinie mehr gefrbt.
                 */
                kaestchen[i][j]++;

                /*
                 * Falls insgesamt vier Randlinien gefrbt sind ...
                 */
                if (kaestchen[i][j] == 4)
                {
                    /*
                     * ... frbe das Kstchen.
                     */
                    zweiteFaerbungX = i;
                    zweiteFaerbungY = j;
                    gefaerbt = true;
                    gibPunkt ();
                }

                /*
                 * Falls kein Kstchen gefrbt wurde ...
                 */
                if (!gefaerbt)
                {
                    /*
                     * ... wechsle zur anderen Spielerin.
                     */
                    naechsteSpielerin = 3 - naechsteSpielerin;
                }

                return pruefeSpielende ();
            }
        }
        /*
         * Wenn die Spielerin nicht dran ist ...
         */ 
        else 
        {
            /*
             * ... melde dies.
             */
            return Ergebnis.NICHT_DRAN;
        }
    }

    /**
     * Gib der Spielerin, die gerade dran ist, einen Punkt.
     */
    private void gibPunkt ()
    {
        if (naechsteSpielerin == 1)
        {
            punkte1 ++;
        }

        if (naechsteSpielerin == 2)
        {
            punkte2 ++;
        }
    }

    /**
     * Prfe ob das Spiel beendet ist und melde, wer gewonnen hat.
     */
    public Ergebnis pruefeSpielende ()
    {
        /*
         * Wenn das Spiel beendet ist ...
         */
        if (punkte1+ punkte2 == x*y)
        {
            /*
             * ... und die erste Spielerin mehr Punkte hat ...
             */
            if (punkte1 > punkte2)
            {
                /*
                 * ... verknde sie als Gewinnerin.
                 */
                gewinnerin = spielerin1;
                return Ergebnis.SPIELERIN1;
            }
            /*
             * Wenn die erste Spielerin mehr Punkte hat ...
             */
            else if (punkte2 > punkte1)
            {
                /*
                 * ... verknde sie als Gewinnerin.
                 */
                gewinnerin = spielerin2;
                return Ergebnis.SPIELERIN2;

            }
            /*
             * ... sonst ...
             */
            else 
            {
                /*
                 * ... verknde ein Unentschieden.
                 */
                return Ergebnis.UNENTSCHIEDEN;
            }
        }
        /*
         * Wenn das Spiel noch nicht zu Ende ist ...
         */
        else 
        {
            /*
             * ... melde dies.
             */
            return Ergebnis.WEITER;
        }
    }

    /**
     * Nenne den Namen der Gewinnerin.
     */
    public String gibGewinnerin ()
    {
        return gewinnerin;
    }

    /**
     * Nenne die X-Koordinate des 1. gefaerbten Kaestchens.
     */
    public int gibErsteFaerbungX ()
    {
        return ersteFaerbungX ;
    }

    /**
     * Nenne die Y-Koordinate des 1. gefaerbten Kaestchens.
     */
    public int gibErsteFaerbungY ()
    {
        return ersteFaerbungY;
    }

    /**
     * Nenne die X-Koordinate des 2. gefaerbten Kaestchens.
     */
    public int gibZweiteFaerbungX ()
    {
        return zweiteFaerbungX;
    }

    /**
     * Nenne die Y-Koordinate des 2. gefaerbten Kaestchens.
     */
    public int gibZweiteFaerbungY ()
    {
        return zweiteFaerbungY;
    }

    /**
     * Nenne den Namen der Spielerin, die als nchste dran ist.
     */
    public String gibNaechsteSpielerin ()
    {
        if (naechsteSpielerin == 1)
        {
            return spielerin1;
        }
        else if (naechsteSpielerin == 2)
        {
            return spielerin2;
        }
        else
        {
            return "";
        }
    }
}
